/*****************************************************************************
 *                                  _   _ ____  _     
 *  Project                     ___| | | |  _ \| |    
 *                             / __| | | | |_) | |    
 *                            | (__| |_| |  _ <| |___ 
 *                             \___|\___/|_| \_\_____|
 *
 * Copyright (C) 1998 Daniel Stenberg and Rafael Sagula
 * ------------------------------------------------------------
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *   MA 02111-1307, USA
 * ------------------------------------------------------------
 * Main author:
 * - Daniel Stenberg <Daniel.Stenberg@sth.frontec.se>
 *
 * 	http://www.fts.frontec.se/~dast/curl/
 *
 * $Source: /home/hardcore/dast/src/curl/RCS/url.c,v $
 * $Revision: 1.14 $
 * $Date: 1998/07/24 11:56:50 $
 * $Author: dast $
 * $State: Exp $
 * $Locker: dast $
 *
 * ------------------------------------------------------------
 * $Log: url.c,v $
 * Revision 1.14  1998/07/24 11:56:50  dast
 * 4.5 fix
 *
 * Revision 1.13  1998/06/06 12:46:13  dast
 * more SSL fix
 * silly non-zero-terminated-buffer bug corrected
 *
 * Revision 1.12  1998/05/27 22:08:02  dast
 * Linas did more SSL fixes
 * -I and -i fixed to work again
 *
 * Revision 1.11  1998/05/16 18:02:33  dast
 * Linas Vepstas' changes
 *
 * Revision 1.10  1998/05/13 17:41:41  dast
 * win32 adjusted
 *
 * Revision 1.7  1998/04/06 14:35:07  dast
 * includes sys/param.h conditionally
 * no longer uses ulong
 *
 * Revision 1.6  1998/04/06 14:01:56  dast
 * -T for HTTP
 * SSL
 *
 * Revision 1.5  1998/03/27 07:38:46  dast
 * binds to a proper port straight away
 * uses ./ instead of / as default
 *
 * Revision 1.4  1998/03/25 16:48:15  dast
 * Fixex GetHost()
 *
 * Revision 1.3  1998/03/23 15:22:33  dast
 * removed all uses of strerror() - they don't seem platform independent
 * enough
 * Added support for CONF_FTPAPPEND
 *
 * Revision 1.2  1998/03/23 10:36:27  dast
 * fixed header
 * added sagulas' changes
 *
 *****************************************************************************/

/*
 * Copyright (c) 1998 Linas Vepstas <linas@linas.org>
 */
#ifdef USE_SSLEAY
/* Copyright (c) 1996 Sampo Kellomaki <sampo@iki.fi>  */
#endif /* USE_SSLEAY */

/* -- WIN32 approved -- */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>

/* you might need to edit the config.h */
#include "config.h"

#include "url.h"

#ifdef WIN32
#include <winsock.h>
#include <time.h>
#include <io.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <signal.h>

#ifdef INCLUDE_PARAM_H
#include <sys/param.h>
#endif
#endif


#ifdef USE_SSLEAY
/* SSLeay stuff usually in /usr/local/ssl/include */
#include "rsa.h"       
#include "crypto.h"
#include "x509.h"
#include "pem.h"
#include "ssl.h"
#include "err.h"

#define CHK_NULL(x) if ((x)==NULL) exit (1)
#define CHK_ERR(err,s) if ((err)==-1) { perror(s); exit(1); }
#define CHK_SSL(err) if ((err)==-1) { ERR_print_errors_fp(stderr); exit(2); }

#endif /* USE_SSLEAY */

/* -- -- */

/* Don't show progress for sizes smaller than: */
#define LEAST_SIZE_PROGRESS BUFSIZE

#define BUFSIZE (2048*5)

/* We don't support header lengths larger than this, in cases where we need to
   append parts to build full lines (i.e longer lines may be fine, but they
   may also not be fine). */
#define HEADERSIZE 256


#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif


/***********************************************************************
 *        config struct
 **********************************************************************/

struct UrlData {
  FILE *out;   /* the fetched file goes here */
  FILE *in;    /* the uploaded file is read from here */
  char *url;   /* what to get */
  char *hostname; /* hostname to contect, as parsed from url */
  unsigned short port; /* which port to use (if non-protocol bind) set
                          CONF_PORT to use this */
  char *proxy; /* if proxy, set it here, set CONF_PROXY to use this */
  unsigned short proxy_port; /* port at which proxy is located  */

  long conf;   /* configure flags */
  char *userpwd;  /* <user:password>, if used */
  char *proxyuserpwd;  /* Proxy <user:password>, if used */
  char *range; /* range, if used. See README for detailed specification on
                  this syntax. */
  char *postfields; /* if POST, set the fields' values here */
  char *referer;
  char *errorbuffer; /* store failure messages in here */
  char *useragent;   /* User-Agent string */

  char *ftpport; /* port to send with the PORT command */

 /* function that stores the output:*/
  size_t (*fwrite)(char *buffer,
                   int size,
                   int nitems,
                   FILE *outstream);

  /* function that reads the input:*/
  size_t (*fread)(char *buffer,
                  int size,
                  int nitems,
                  FILE *outstream);

  long timeout; /* in seconds, 0 means no timeout */
  long infilesize; /* size of file to upload, -1 means unknown */

  /* fields only set and used within _urlget() */
  int firstsocket;     /* the main socket to use */
  int secondarysocket; /* for i.e ftp transfers */

  char buffer[BUFSIZE+1]; /* buffer with size BUFSIZE */

  int current_speed;  /* the ProgressShow() funcion sets this */

  int low_speed_limit; /* bytes/second */
  int low_speed_time;  /* number of seconds */

  int resume_from;    /* continue [ftp] transfer from here */

  short    use_ssl;   /* use ssl encrypted communications */

#ifdef USE_SSLEAY
  SSL_CTX* ctx;
  SSL*     ssl;
  X509*    server_cert;
#endif /* USE_SSLEAY */

};

static int sendf(int fd, struct UrlData *, char *fmt, ...);
static void infof(struct UrlData *, char *fmt, ...);
static void failf(struct UrlData *, char *fmt, ...);


/***********************************************************************
 * Start with some silly functions to make win32-systems survive
 ***********************************************************************/
#ifdef WIN32
static void cleanup(void)
{
  WSACleanup();
}

static int init(void)
{
  WORD wVersionRequested;  
  WSADATA wsaData; 
  int err; 
  wVersionRequested = MAKEWORD(1, 1); 
    
  err = WSAStartup(wVersionRequested, &wsaData); 
    
  if (err != 0) 
    /* Tell the user that we couldn't find a useable */ 
    /* winsock.dll.     */ 
    return 1; 
    
  /* Confirm that the Windows Sockets DLL supports 1.1.*/ 
  /* Note that if the DLL supports versions greater */ 
  /* than 1.1 in addition to 1.1, it will still return */ 
  /* 1.1 in wVersion since that is the version we */ 
  /* requested. */ 
    
  if ( LOBYTE( wsaData.wVersion ) != 1 || 
       HIBYTE( wsaData.wVersion ) != 1 ) { 
    /* Tell the user that we couldn't find a useable */ 

    /* winsock.dll. */ 
    WSACleanup(); 
    return 1; 
  }
  return 0;
}
/* The Windows Sockets DLL is acceptable. Proceed. */ 
#else
static int init(void) { return 0; }
static void cleanup(void) {}
#endif

static UrgError _urlget(struct UrlData *data);

char *getmyhost(void)
{
  static char myhost[256];
#ifdef WIN32
  gethostname(myhost, 256);
#else
  struct utsname ugnm;

  if (uname(&ugnm) < 0)
    return "localhost";

  (void) strncpy(myhost, ugnm.nodename, 255);
  myhost[255] = '\0';
#endif
#if 0
  FILE *host;
  /* this is a reaaaaaaally lame and non-portable way of doing this you
     know... */

 host = popen("hostname", "r");
  if(host) {
    char *ptr;
    fgets(myhost, 256, host);
    
    ptr= strchr(myhost, '\n');
    if(ptr)
      *ptr = '\0';
    fclose(host);
  }
  else {
    strcpy(myhost, "127.0.0.1");
  }
#endif
  return myhost;
}

#ifndef WIN32
/* -- if2ip() -- */
#include <netdb.h>

#if USE_SOCKIO
#include <sys/sockio.h>
#endif

#define SYS_ERROR -1

char *if2ip(char *interface)
{
  int dummy;
  char *ip=NULL;
  
  if(!interface)
    return NULL;

  dummy = socket(AF_INET, SOCK_STREAM, 0);
  if (SYS_ERROR == dummy) {
    return NULL;
  }
  else {
    struct ifreq req;
    memset(&req, 0, sizeof(req));
    strcpy(req.ifr_name, interface);
    req.ifr_addr.sa_family = AF_INET;
    if (SYS_ERROR == ioctl(dummy, SIOCGIFADDR, &req, sizeof(req))) {
      return NULL;
    }
    else {
      struct in_addr in;
      struct ifreq ifr;
      struct sockaddr_in *s = (struct sockaddr_in *)&req.ifr_dstaddr;
      memcpy(&in, &(s->sin_addr.s_addr), sizeof(in));
      ip = (char *)strdup(inet_ntoa(in));
    }
    close(dummy);
  }
  return ip;
}

/* -- end of if2ip() -- */
#else
#define if2ip(x) NULL
#endif

void urlfree(struct UrlData *data)
{
  /* close possibly still open sockets */
  if(-1 != data->secondarysocket)
    sclose(data->secondarysocket);
  if(-1 != data->firstsocket)
    sclose(data->firstsocket);

#ifdef USE_SSLEAY
  if (data->use_ssl) {
    SSL_free (data->ssl);
    SSL_CTX_free (data->ctx);
  }
#endif /* USE_SSLEAY */

  free(data);

  /* winsock crap cleanup */
  cleanup();
}

UrgError urlget(UrgTag tag, ...)
{
  va_list arg;
  void *param;
  UrgError res;

  struct UrlData *data;

  /* this is for the lame win32 socket crap */
  if(init())
    return URG_FAILED_INIT;

  data = malloc(sizeof(struct UrlData));
  if(data) {

    memset(data, 0, sizeof(struct UrlData));

    /* Let's set some default values: */
    data->out = stdout; /* default output to stdout */
    data->in  = stdin;  /* default input from stdin */
    data->firstsocket = -1; /* no file descriptor */
    data->secondarysocket = -1; /* no file descriptor */

    /* use fwrite as default function to store output */
    data->fwrite = (size_t (*)(char *, int, int, FILE *))fwrite;

    /* use fread as default function to read input */
    data->fread = (size_t (*)(char *, int, int, FILE *))fread;

    data->infilesize = -1; /* we don't know any size */

    data->port = 0;
    data->proxy_port = 1080;
    data->url = NULL;
    data->hostname = NULL;
    data->proxy = NULL;
    data->referer = NULL;
    data->useragent = NULL;
    data->use_ssl = 0;

    data->current_speed = -1; /* init to negative == impossible */

    va_start(arg, tag);

    while(tag != URGTAG_DONE) {
      /*      printf("tag: %d\n", tag); */
      param = va_arg(arg, void *);
      
      switch(tag) {
      case URGTAG_ERRORBUFFER:
        data->errorbuffer = (char *)param;
        break;
      case URGTAG_FILE:
        data->out = (FILE *)param;
        break;
      case URGTAG_FTPPORT:
        data->ftpport = (char *)param;
        break;
      case URGTAG_INFILE:
        data->in = (FILE *)param;
        break;
      case URGTAG_INFILESIZE:
        data->infilesize = (long)param;
        break;
      case URGTAG_LOW_SPEED_LIMIT:
	data->low_speed_limit=(int)param;
	break;
      case URGTAG_LOW_SPEED_TIME:
	data->low_speed_time=(int)param;
	break;
      case URGTAG_URL:
        data->url = (char *)param;
        break;
      case URGTAG_PORT:
        /* this typecast is used to fool the compiler to NOT warn for a
           "cast from pointer to integer of different size" */
        data->port = (unsigned short)((unsigned int)param);
        break;
      case URGTAG_POSTFIELDS:
        data->postfields = (char *)param;
        break;
      case URGTAG_REFERER:
        data->referer = (char *)param;
        break;
      case URGTAG_PROXY:
        data->proxy = (char *)param;
        break;
      case URGTAG_FLAGS:
        data->conf = (long)param;
        break;
      case URGTAG_TIMEOUT:
        data->timeout = (long)param;
        break;
      case URGTAG_USERAGENT:
        data->useragent = (char *)param;
        break;
      case URGTAG_USERPWD:
        data->userpwd = (char *)param;
        break;
      case URGTAG_PROXYUSERPWD:
        data->proxyuserpwd = (char *)param;
        break;
      case URGTAG_RANGE:
        data->range = (char *)param;
        break;
      case URGTAG_RESUME_FROM:
	data->resume_from = (int)param;
	break;
      case URGTAG_WRITEFUNCTION:
        data->fwrite = (size_t (*)(char *, int, int, FILE *))param;
        break;
      case URGTAG_READFUNCTION:
        data->fread = (size_t (*)(char *, int, int, FILE *))param;
        break;
      case URGTAG_DONE: /* done with the parsing, fall through */
        continue;
      default:
        /* unknown tag and its companion, just ignore: */
        break;
      }
      tag = va_arg(arg, UrgTag);
    }

    va_end(arg);

    res = _urlget(data); /* fetch the URL please */

  }
  else
    res = URG_FAILED_INIT; /* failed */

  /* cleanup */
  urlfree(data);

  return res;
}


/* ---- Base64 Encoding --- */
static char table64[]=
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  
void base64Encode(char *intext, char *output)
{
  unsigned char ibuf[3];
  unsigned char obuf[4];
  int i;
  int inputparts;

  while(*intext) {
    for (i = inputparts = 0; i < 3; i++) { 
      if(*intext) {
        inputparts++;
        ibuf[i] = *intext;
        intext++;
      }
      else
        ibuf[i] = 0;
    }
                       
    obuf [0] = (ibuf [0] & 0xFC) >> 2;
    obuf [1] = ((ibuf [0] & 0x03) << 4) | ((ibuf [1] & 0xF0) >> 4);
    obuf [2] = ((ibuf [1] & 0x0F) << 2) | ((ibuf [2] & 0xC0) >> 6);
    obuf [3] = ibuf [2] & 0x3F;

    switch(inputparts) {
    case 1: /* only one byte read */
      sprintf(output, "%c%c==", 
              table64[obuf[0]],
              table64[obuf[1]]);
      break;
    case 2: /* two bytes read */
      sprintf(output, "%c%c%c=", 
              table64[obuf[0]],
              table64[obuf[1]],
              table64[obuf[2]]);
      break;
    default:
      sprintf(output, "%c%c%c%c", 
              table64[obuf[0]],
              table64[obuf[1]],
              table64[obuf[2]],
              table64[obuf[3]] );
      break;
    }
    output += 4;
  }
  *output=0;
}
/* ---- End of Base64 Encoding ---- */


/* --- start of progress routines --- */
int progressmax=-1;

void ProgressInit(struct UrlData *data, int max)
{
  static char beenhere=0;

  if(beenhere || (data->conf&CONF_NOPROGRESS))
    return;
  beenhere=1;
  progressmax = max;
  if(-1 == max)
    return;
  if(progressmax <= LEAST_SIZE_PROGRESS) {
    progressmax = -1; /* disable */
    return;
  }

  fprintf(stderr, "  %%   Received   Total  Speed   Time left  Total   Curr.Speed\n");
}

void time2str(char *r, int t)
{
  int h = (t/3600);
  int m = (t-(h*3600))/60;
  int s = (t-(h*3600)-(m*60));
  sprintf(r,"%3d:%02d:%02d",h,m,s);
}

void ProgressShow(struct UrlData *data,
                  int point, int start, int now)
{
  static time_t lastshow;
  int spent;
  int speed;

#define CURR_TIME 5

  static int speeder[ CURR_TIME ];
  static int speeder_c=0;

  int nowindex = speeder_c% CURR_TIME;
  int checkindex;
  int count;

  if((point != progressmax) && (lastshow == now))
    return; /* never update this more than once a second if the end isn't 
               reached */

  spent = now-start;
  speed = point/(spent?spent:1);
  if(!speed)
    speed=1;

  /* point is where we are right now */
  speeder[ nowindex ] = point;
  speeder_c++; /* increase */
  count = (speeder_c>=CURR_TIME)?CURR_TIME:speeder_c;
  checkindex = (speeder_c>=CURR_TIME)?speeder_c%CURR_TIME:0;

  /* find out the average speed the last CURR_TIME seconds */
  data->current_speed = (speeder[nowindex]-speeder[checkindex])/count;

#if 0
  printf("NOW %d(%d) THEN %d(%d) DIFF %d COUNT %d\n",
	 speeder[nowindex], nowindex,
	 speeder[checkindex], checkindex,
	 data->current_speed, count);
#endif

  if(data->conf&CONF_NOPROGRESS)
    return;

  if(-1 != progressmax) {
    char left[20],estim[20];
    int estimate = progressmax/speed;
    
    time2str(left,estimate-spent); 
    time2str(estim,estimate);

    fprintf(stderr, "\r%3d %8d  %8d %6d %s %s %6d   ",
            point*100/progressmax, point, progressmax,
            speed, left, estim, data->current_speed);
  }
  else
    fprintf(stderr, "\r%d bytes received in %d seconds (%d bytes/sec)",
            point, spent, speed);

  lastshow = now;
}

void ProgressEnd(struct UrlData *data)
{
  if(data->conf&CONF_NOPROGRESS)
    return;
  fputs("\n", stderr);
}

/* --- end of progress routines --- */


/* --- resolve name or IP-number --- */

char *MakeIP(unsigned long num)
{
#ifdef HAVE_INET_NTOA
  struct in_addr in;

  in.s_addr = htonl(num);
  return (inet_ntoa(in));
#else
  static char addr[MIDBUFFER];
  unsigned char *paddr;

  num = htonl(num);  /* htonl() added to avoid endian probs */
  paddr = (unsigned char *)&num;
  sprintf(addr, "%u.%u.%u.%u", paddr[0], paddr[1], paddr[2], paddr[3]);
  return (addr);
#endif
}

/* Stolen from Dancer source code, written by
   Bjorn Reese <breese@imada.ou.dk> */
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif
struct hostent *GetHost(struct UrlData *data, char *hostname)
{
  struct hostent *h = NULL;
  unsigned long in;
  static struct hostent he;
  static char name[MAXHOSTNAMELEN];
  static char *addrlist[2];
  static struct in_addr addrentry;

  if ( (in=inet_addr(hostname)) != INADDR_NONE ) {
    addrentry.s_addr = in;
    addrlist[0] = (char *)&addrentry;
    addrlist[1] = NULL;
    he.h_name = strncpy(name, MakeIP(ntohl(in)), MAXHOSTNAMELEN);
    he.h_addrtype = AF_INET;
    he.h_length = sizeof(struct in_addr);
    he.h_addr_list = addrlist;
    h = &he;
  } else if ( (h=gethostbyname(hostname)) == NULL ) {
    infof(data, "gethostbyname(2) failed for %s\n", hostname);
  }
  return (h);
}

/* --- parse FTP server responses --- */

static int GetLastResponse(int sockfd, char *buf, struct UrlData *data)
{
  int nread;
  int read_rc=1;
  char *ptr;
  do {
    ptr=buf;

    /* get us a full line, terminated with a newline */
    for(nread=0;
        (nread<BUFSIZE) && read_rc;
        nread++, ptr++) {
#ifndef USE_SSLEAY
      read_rc = sread(sockfd, ptr, 1);
#else
      if (data->use_ssl) {
        read_rc = SSL_read(data->ssl, ptr, 1);
      } else {
        read_rc = sread(sockfd, ptr, 1);
      }
#endif /* USE_SSLEAY */
      if (*ptr == '\n')
        break;
    }
    *ptr=0; /* zero terminate */

    if(data->conf & CONF_VERBOSE) {
      fputs("< ", stderr);
      fwrite(buf, 1, nread, stderr);
      fputs("\n", stderr);
    }
  } while(read_rc && (nread>3) && ('-'==buf[3]));
  return nread;
}


static UrgError speedcheck(struct UrlData *data,
			   int bytecount,
			   time_t now)
{
  static time_t keeps_speed;

  if((data->current_speed >= 0) &&
     data->low_speed_time &&
     keeps_speed &&
     (data->current_speed < data->low_speed_limit)) {

    /* We are now below the "low speed limit". If we are below it
       for "low speed time" seconds we consider that enough reason
       to abort the download. */
    
    if( (now - keeps_speed) > data->low_speed_time) {
      /* we have been this slow for long enough, now die */
      failf(data,
	    "Operation too slow. "
	    "Less than %d bytes/sec transfered the last %d seconds",
	    data->low_speed_limit,
	    data->low_speed_time);
      return URG_OPERATION_TIMEOUTED;
    }
  }
  else {
    /* we keep up the required speed all right */
    keeps_speed = now;
  }
#if 0
  if (data->conf & CONF_VERBOSE) {
    fprintf(stderr, "* %d > %d\n",
	    data->current_speed, data->low_speed_limit);
  }
#endif
  return URG_OK;
}

/* --- upload a stream to a socket --- */

static UrgError Upload(struct UrlData *data,
                       int sockfd,
                       int size,
                       long *bytecountp)
{
  fd_set writefd;
  fd_set keepfd;
  struct timeval interval;
  bool keepon=TRUE;
  char *buf = data->buffer;
  size_t nread;
  long bytecount=0;
  time_t start = time(NULL);
  time_t now;
  UrgError urg;
      
  /* timeout every X second
     - makes a better progressmeter (i.e even when no data is sent, the
       meter can be updated and reflect reality)
     - allows removal of the alarm() crap
     - variable timeout is easier
     */
            
  myalarm(0); /* switch off the alarm-style timeout */

  FD_ZERO(&writefd); /* clear it */
  FD_SET(sockfd, &writefd);
      
  keepfd = writefd;

  while(keepon) {
    size_t bytes_written = 0;

    writefd = keepfd; /* set this every lap in the loop */
    interval.tv_sec = 2;
    interval.tv_usec = 0;

    switch(select(sockfd+1, NULL, &writefd, NULL, &interval)) {
    case -1: /* error, stop writing */
      keepon=FALSE;
      continue;
    case 0: /* timeout */
      break;
    default: /* write! */
      nread = data->fread(buf, 1, BUFSIZE, data->in);
      bytecount += nread;

      if (nread==0) {
        /* done */
        keepon = FALSE; 
        break;
      }

      /* write to socket */
#ifndef USE_SSLEAY
      bytes_written = swrite(sockfd, buf, nread);
#else
      if (data->use_ssl) {
        bytes_written = SSL_write(data->ssl, buf, nread);
      } else {
        bytes_written = swrite(sockfd, buf, nread);
      }
#endif /* USE_SSLEAY */
      if(nread != bytes_written) {
        failf(data, "Failed uploading file");
        return URG_FTP_WRITE_ERROR;
      }
    }
    now = time(NULL);
    ProgressShow(data, bytecount, start, now);
    urg=speedcheck(data, bytecount, now);
    if(urg)
      return urg;
    if(data->timeout && ((now-start)>data->timeout)) {
      failf(data, "Upload timed out with %d bytes sent", bytecount);
      return URG_OPERATION_TIMEOUTED;
    }

  }
  *bytecountp = bytecount;

  return URG_OK;
}

/* --- download a stream from a socket --- */

  /* This features optional (HTTP) header parsing to
   - show the header
   - find the document size
   */

static
UrgError Download(struct UrlData *data,
                  int sockfd, /* socket to read from */
                  int size, /* -1 if unknown at this point */
                  bool getheader, /* TRUE if header parsing is wanted */
                  long *bytecountp /* return number of bytes read */
                  )
{
  char *buf = data->buffer;
  size_t nread;
  int bytecount=0;
  time_t start=time(NULL);
  time_t now;
  bool header=TRUE;

  char headerbuff[HEADERSIZE];
  char *hbufp=headerbuff;
  int hbuflen=0;
  char *str;

  /* for the low speed checks: */
  UrgError urg;

  myalarm(0); /* switch off the alarm-style timeout */

  if(!getheader) {
    header=FALSE;
    ProgressInit(data, size);
  }
  {
    fd_set readfd;
    fd_set keepfd;
    struct timeval interval;
    bool keepon=TRUE;
      
    /* timeout every X second
       - makes a better progressmeter (i.e even when no data is read, the
         meter can be updated and reflect reality)
       - allows removal of the alarm() crap
       - variable timeout is easier
       */
            
    FD_ZERO(&readfd); /* clear it */
    FD_SET(sockfd, &readfd);
      
    keepfd = readfd;

    while(keepon) {
      readfd = keepfd; /* set this every lap in the loop */
      interval.tv_sec = 2;
      interval.tv_usec = 0;

      switch(select(sockfd+1, &readfd, NULL, NULL, &interval)) {
      case -1: /* error, stop reading */
        keepon=FALSE;
        continue;
      case 0: /* timeout */
        break;
      default: /* read! */
#ifndef USE_SSLEAY
        nread = sread(sockfd, buf, BUFSIZE-1);
#else
        if (data->use_ssl) {
          nread = SSL_read(data->ssl, buf, BUFSIZE-1);
        } else {
          nread = sread(sockfd, buf, BUFSIZE-1);
        }
#endif /* USE_SSLEAY */

        /* NULL terminate, allowing string ops to be used */
        if (0 <= nread)
          buf[nread] = 0;

        /* if we receive 0 here, the server closed the connection and we
           bail out from this! */
        if (nread==0) {
          keepon=FALSE;
          break;
        }

        str = buf; /* Default buffer to use when we write the buffer, it may
                      be changed in the flow below before the actual storing
                      is done. */

        /* Since this is a two-state thing, we check if we are parsing
           headers at the moment or not. */

        if(header) {
          /* we are in parse-the-header-mode */
          int code;

          /* Search for end of line */
          /* str = strchr(buf, '\n'); */

          if(!str) {
            /* We didn't find end of line, make next read buffer get appended
               to the end of this buffer and continue. */
            strcpy(hbufp, buf);
            hbufp += strlen(buf);
            hbuflen += strlen(buf);
            break;
          }
          else {
            do {
              /* Each \r should be followed by a \n, but we do it nicely
                 and check for it anyway */
              if('\n' == *str)
                str++;

              /* in case we got a buffer to append */
              if(hbufp != headerbuff) {
                /* check that we can append this without overflow */
                if(hbuflen + strlen(str) > HEADERSIZE)
                  return URG_READ_ERROR;

                /* reset the variables */
                strcpy(hbufp, str); /* copy the end of the line we want */
                hbufp = headerbuff; /* reset */
                hbuflen=0;
              }
              else
                hbufp=str; /* point out our line to scan */

              /* we now have a full line that hbufp points to */
              if(('\n' == *hbufp) ||
		 ('\r' == *hbufp) ) {
                /* Zero-length line means end of header! */
                if(-1 != size) /* if known */
                  size += bytecount; /* we append the already read size */
		if ('\r' == *str)
		  str++; /* pass the \r byte */
		if('\n' == *str)
		  str++; /* pass the \n byte */

                ProgressInit(data, size); /* init progress meter */
                header=FALSE; /* no more header to parse! */

                /* now, only output this if the header AND body are requested: */
                if((data->conf&(CONF_HEADER|CONF_NOBODY)) == CONF_HEADER)
                  data->fwrite(hbufp, 1, str-hbufp, data->out);
                break;
              }
              else if(data->conf & CONF_FAILONERROR) {
                /* If we have been told to fail hard on HTTP-errors,
                   here is the check for that: */
                if (sscanf(hbufp, "HTTP/1.%*c %3d", &code)) {
                  /* 404 -> URL not found! */
                  if(code >= 300) {
                    /* serious error, go home! */
                    failf(data, "The requested file was not found");
                    return URG_HTTP_NOT_FOUND;
                  }
                }
              }
              /* check for Content-Length: header lines to get size */
              sscanf(hbufp, "Content-%*cength: %d", &size);

              /* find the next line in our buffer: */
              str=strchr(str, '\n');

              if(data->conf & CONF_HEADER)
                 data->fwrite(hbufp, 1, str-hbufp+1, data->out);

              /* reset hbufp pointer */
              hbufp = headerbuff;
            } while (str);

            /* We might have reached the end of the header part here, but
               there might be a non-header part left in the end of the read
               buffer. */

            if(!header) { /* the next token and forward is not part of
                             the header! */

              /* we subtract the header size from the buffer */
              nread -= (str-buf);
            }
          }
        }

        /* This is not an 'else if' since it may be a rest from the header
           parsing, where the beginning of the buffer is headers and the end
           is non-headers. */
        if(!header && (nread>0)) {
          bytecount += nread;
          data->fwrite(str, 1, nread, data->out);
#ifdef CHECK_THIS_OUT
	  if(nread != data->fwrite(str, 1, nread, data->out)) {
            failf(data, "Failed writing output");
            return URG_WRITE_ERROR;
          }
#endif
        }
        break;
      }
      now = time(NULL);
      if(!header) {
        ProgressShow(data, bytecount, start, now);
      }
      urg=speedcheck(data, bytecount, now);
      if(urg)
	return urg;

      if(data->timeout && ((now-start)>data->timeout)) {
        failf(data, "Operation timed out with %d out of %d bytes received",
              bytecount, size);
        return URG_OPERATION_TIMEOUTED;
      }
    }
  }
  *bytecountp = bytecount;
  return URG_OK;
}

UrgError AllowServerConnect(struct UrlData *data,
			    int sock)
{
  fd_set rdset;
  struct timeval dt;
  
  FD_ZERO(&rdset);

  FD_SET(sock, &rdset);

  /* we give the server 10 seconds to connect to us */
  dt.tv_sec = 10;
  dt.tv_usec = 0;

  switch ( select(sock+1, &rdset, NULL, NULL, &dt)) {
  case -1: /* error */
    /* let's die here */
    failf(data, "Error while waiting for server connect");
    return URG_FTP_PORT_FAILED;
  case 0:  /* timeout */
    /* let's die here */
    failf(data, "Timeout while waiting for server connect");
    return URG_FTP_PORT_FAILED;
  default:
    /* we have received data here */
    {
      int s;
      size_t size = sizeof(struct sockaddr_in);
      struct sockaddr_in add;

      getsockname(sock, (struct sockaddr *) &add, &size);
      s=accept(sock, (struct sockaddr *) &add, &size);

      if( -1 == s) {
	/* DIE! */
	failf(data, "Error accept()ing server connect");
	return URG_FTP_PORT_FAILED;
      }
      infof(data, "Connection accepted from server\n");

      data->secondarysocket = s;
    }
    break;
  }
  return URG_OK;
}

/* ====================================================== */
void 
UrgSSLConnect (struct UrlData *data)
{
    int err;
    char * str;

    /* mark this is being ssl enabled from here on out. */
    data->use_ssl = 1;

#ifdef USE_SSLEAY
    /* Lets get nice error messages */
    SSL_load_error_strings();

    /* Setup all the global SSL stuff */
    SSLeay_add_ssl_algorithms();
    data->ctx = SSL_CTX_new(SSLv23_client_method());

    /* Lets make a SSL structure */
    data->ssl = SSL_new (data->ctx);
    SSL_set_connect_state (data->ssl);

    data->server_cert = 0x0;

    /* pass the raw socket into the SSL layers */
    SSL_set_fd (data->ssl, data->firstsocket);
    err = SSL_connect (data->ssl);               CHK_SSL(err);

    infof (data, "SSL connection using %s\n", SSL_get_cipher (data->ssl));
  
    /* Get server's certificate (note: beware of dynamic allocation) - opt */
    /* major serious hack alert -- we should check certificates
     * to authenticate the server; otherwise we risk man-in-the-middle
     * attack
     */

    data->server_cert = SSL_get_peer_certificate (data->ssl);
    CHK_NULL(data->server_cert);
    infof (data, "Server certificate:\n");
  
    str = X509_NAME_oneline (X509_get_subject_name (data->server_cert), NULL, 0);
    CHK_NULL(str);
    infof (data, "\t subject: %s\n", str);
    Free (str);

    str = X509_NAME_oneline (X509_get_issuer_name  (data->server_cert), NULL, 0);
    CHK_NULL(str);
    infof (data, "\t issuer: %s\n", str);
    Free (str);

    /* We could do all sorts of certificate verification stuff here before
       deallocating the certificate. */

    X509_free (data->server_cert);
#endif /* USE_SSLEAY */

}

#ifndef WIN32
void alarmfunc(int signal)
{
  return;
}
#endif

/* ====================================================== */
/*
 * urlget <url>
 * (result put on stdout)
 *
 * <url> ::= <proto> "://" <host> [ ":" <port> ] "/" <path>
 *
 * <proto> = "HTTP" | "HTTPS" | "GOPHER" | "FTP"
 *
 * When FTP:
 *
 * <host> ::= [ <user> ":" <password> "@" ] <host>
 */

static UrgError _urlget(struct UrlData *data)
{
  struct hostent *hp=NULL;
  struct sockaddr_in serv_addr;
  char *buf;
  char proto[64];
  char gname[256]="default.com";
  char *name;
  char path[512]="/";
  char *ppath, *tmp;
  long bytecount;
  time_t now;
  char ftpuser[128]="";
  char ftppasswd[128]="";
  char proxyuser[128]="";
  char proxypasswd[128]="";
  char proxyuserpwd[512];
  char uagent[512];
  size_t nread;
  UrgError result;

  /* Temporary kludgey fix until I replace properly in the source: */
  char *proxy = data->proxy; /* if proxy, set it here, set CONF_PROXY to use
                                this */
  long conf = data->conf;   /* configure flags */


  /* for the ftp PORT mode */
  int portsock;

  int saveport = data->port;  /* work arount the mess involving proxies & port numbers */

  buf = data->buffer; /* this is our buffer */

#ifndef WIN32
  signal(SIGALRM, alarmfunc);
#endif

  /* Parse <url> */
  /* We need to parse the url, even when using the proxy, because
   * we will need the hostname and port in case we are trying
   * to SSL connect through the proxy -- and we don't know if we
   * will need to use SSL until we parse the url ...
   */
  if (2 > sscanf(data->url, "%64[^\n:]://%256[^\n/]%512[^\n]",
		   proto, gname, path)) {
      
    /* badly formatted, let's try the browser-style _without_ 'http://' */
    if((1 > sscanf(data->url, "%256[^\n/]%512[^\n]", gname, path)) ) {
      failf(data, "<url> malformed");
      return URG_URL_MALFORMAT;
    }
    if(strnequal(gname, "FTP", 3))
      strcpy(proto, "FTP");
    else if(strnequal(gname, "GOPHER", 6))
      strcpy(proto, "GOPHER");
#ifdef USE_SSLEAY
    else if(strnequal(gname, "HTTPS", 5))
      strcpy(proto, "HTTPS");
#endif /* USE_SSLEAY */
    else
      strcpy(proto, "HTTP");
  }

  if(conf & CONF_USERPWD) {
    if(':' != *data->userpwd) {
      if((1 <= sscanf(data->userpwd, "%127[^:]:%127s",
                      ftpuser, ftppasswd))) {
        ; /* just continue */
      }
    }
    if(!ftpuser[0]) {
      failf(data, "USER malformat: user name can't be zero length");
      return URG_MALFORMAT_USER;
    }
  }

  if(conf & CONF_PROXYUSERPWD) {
    if(':' != *data->proxyuserpwd) {
      if((1 <= sscanf(data->proxyuserpwd, "%127[^:]:%127s",
                      proxyuser, proxypasswd))) {
        ; /* just continue */
      }
    }
    if(!proxyuser[0]) {
      failf(data, " Proxy USER malformat: user name can't be zero length");
      return URG_MALFORMAT_USER;
    }
  }

  name = gname;
  ppath = path;
  data->hostname = name;

  /* 
   * Save the value of the proxy port (if specified) now, because
   * the parsing below will destroy the port value.  Basically,
   * we just copy the port number if the user specified both 
   * the -x and -p flags.
   */
  if (conf & CONF_PROXY) {
    if (conf & CONF_PORT) {
        data->proxy_port = data->port;
    } else {
        data->proxy_port = 1080;
    }
  }


  if(data->timeout) {
    /* We set the timeout on the connection/resolving phase first, separately
       from the download/upload part to allow a maximum time on everything */
    myalarm(data->timeout); /* this sends a signal when the timeout fires
			       off, and that will abort system calls */
  }

  /*
   * Hmm, if we are using a proxy, then we can skip the GOPHER and the
   * FTP steps, although we cannot skip the HTTPS step (since the proxy
   * works differently, depending on whether its SSL or not).
   */

  if (strequal(proto, "HTTP")) {
    data->port = 80;
    conf |= CONF_HTTP;
  }
  else if (strequal(proto, "HTTPS")) {
#ifdef USE_SSLEAY
    data->port = 443;
    conf |= CONF_HTTP;
    conf |= CONF_HTTPS;
#else /* USE_SSLEAY */
    failf(data, "SSL is disabled, https: not supported!");
    return URG_UNSUPPORTED_PROTOCOL;
#endif /* !USE_SSLEAY */
  }
  else if (strequal(proto, "GOPHER")) {
    data->port = 70;
    /* Skip /<item-type>/ in path if present */
    if (isdigit(path[1])) {
      ppath = strchr(&path[1], '/');
      if (ppath == NULL)
	ppath = path;
      }
    conf |= CONF_GOPHER;
  }
  else if(strequal(proto, "FTP")) {

    data->port = 21;
    conf |= CONF_FTP;

    /* 3.2-fix, why did we remove this? */
    ppath++; /* don't include the initial slash */

  }
  else {
    failf(data, "Unsupported protocol: %s", proto);
    return URG_UNSUPPORTED_PROTOCOL;
  }

  if(conf & CONF_NETRC) {
    if(ParseNetrc(data->hostname, ftpuser, ftppasswd)) {
      infof(data, "Couldn't find host %s in the .netrc file, using defaults",
            data->hostname);
    }
    /* weather we failed or not, we don't know which fields that were filled
       in anyway */
    if(!ftpuser[0])
      strcpy(ftpuser, "anonymous");
    if(!ftppasswd[0])
      strcpy(ftppasswd, "curl_by_Daniel.Stenberg@sth.frontec.se");
    if(conf & CONF_HTTP) {
      conf |= CONF_USERPWD;
    }
  }
  else if(!(conf & CONF_USERPWD) &&
	  (conf & (CONF_FTP|CONF_HTTP)) ) {
    /* This is a FTP or HTTP URL, and we haven't got the user+password in
       the extra parameter, we will now try to extract the possible
       user+password pair in a string like:
       ftp://user:password@ftp.my.site:8021/README */
    char *ptr;
    if(':' == *name) {
      failf(data, "URL malformat: user can't be zero length");
      return URG_URL_MALFORMAT_USER;
    }
    if((1 <= sscanf(name, "%127[^:]:%127[^@]",
		    ftpuser, ftppasswd)) && (ptr=strchr(name, '@'))) {
      name = ++ptr;
      conf |= CONF_USERPWD;
    }
    else {
      strcpy(ftpuser, "anonymous");
      strcpy(ftppasswd, "curl_by_Daniel.Stenberg@sth.frontec.se");
    }
  }
  
   /* if a proxy is not specified, but a port number is,
    * override the default port number.  But allow anything
    * embedded in the URL to override the -p flag. 
    */
   if ((!(conf & CONF_PROXY)) && (conf & CONF_PORT)) {
      data->port = saveport;
   }

    /* extract the port from the URL, if it is there, thus overriding
     * any defaults that might have been set above. */
    tmp = strchr(name, ':');
    if (tmp) {
      *tmp++ = '\0';
      data->port = atoi(tmp);
    }

  if (conf & CONF_PROXY) {
    if(!(hp = GetHost(data, proxy))) {
      failf(data, "Couldn't resolv proxy '%s'", proxy);
      return URG_COULDNT_RESOLVE_PROXY;
    }
  } else {
    if(!(hp = GetHost(data, name))) {
      failf(data, "Couldn't resolv host '%s'", name);
      return URG_COULDNT_RESOLVE_HOST;
    }
  }

  data->firstsocket = socket(AF_INET, SOCK_STREAM, 0);

  memset((char *) &serv_addr, '\0', sizeof(serv_addr));
  memcpy((char *)&(serv_addr.sin_addr), hp->h_addr, hp->h_length);
  serv_addr.sin_family = hp->h_addrtype;
  if (conf & CONF_PROXY) {
    serv_addr.sin_port = htons(data->proxy_port);
  } else {
    serv_addr.sin_port = htons(data->port);
  }
  
  if (connect(data->firstsocket, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    switch(errno) {
#ifndef WIN32
      /* this should be made nicer */
    case ECONNREFUSED:
      failf(data, "Connection refused");
      break;
    case EINTR:
      failf(data, "Connection timeouted");
      break;
#endif
    default:
      failf(data, "Can't connect to server: %d", errno);
      break;
    }
    return URG_COULDNT_CONNECT;
  }

  if(conf & CONF_PROXYUSERPWD) {
    char authorization[512];
    sprintf(proxyuserpwd, "%s:%s", proxyuser, proxypasswd);
    base64Encode(proxyuserpwd, authorization);
    sprintf(proxyuserpwd,
            "Proxy-authorization: Basic %s\015\012",
            authorization);
  }
  if(conf & (CONF_HTTPS|CONF_HTTP)) {
    if(data->useragent) {
      sprintf(uagent, "User-Agent: %s\015\012", data->useragent);
    }
  }


  /* If we are not using a proxy and we want a secure connection,
   * perform SSL initialization & connection now.
   * If using a proxy with https, then we must tell the proxy to CONNECT
   * us to the host we want to talk to.  Only after the connect
   * has occured, can we start talking SSL
   */
   if (conf & CONF_HTTPS) {
     if (conf & CONF_PROXY) {
        char buffy[512];
        char * ptr = buffy;
        int read_rc = 1;
        int crnl = 0;

        /* OK, now send the connect statment */
        sendf(data->firstsocket, data,
              "CONNECT %s:%d HTTP/1.0\015\012"
              "%s"
              "%s"
              "\015\012",
              data->hostname, data->port,
              (conf&CONF_PROXYUSERPWD)?proxyuserpwd:"",
              (data->useragent?uagent:"")
              );

        /* wait for the proxy to send us a HTTP/1.0 200 OK header */
        *ptr = 0x0;
        while (1) {
          read_rc = sread(data->firstsocket, ptr, 1);
          if (0>= read_rc) break;
          /* we'll know its a blank line if thre are two new-lines in a row,
           * with only carriage returns getting in the way */
          if ('\n' == *ptr) crnl ++;
          if (('\r' != *ptr) && ('\n' != *ptr)) crnl = 0;
          if (2 <= crnl) break;
          ptr ++;
          *ptr = 0x0;
        }
        ptr ++;
        *ptr = 0x0;

        if (2 > crnl) {
           /* oops we exited the loop without seeing the 'blank line' */
           /* that's an error condition. */
	  failf (data,
		 "Error: Read error while chatting with proxy \n"
		 "Received:  %s", buffy);
	  return URG_READ_ERROR;
        }
        if (data->conf & CONF_VERBOSE)
             fprintf(stderr, "< %s", buffy);

        /* now scan for an HTTP 200 return code */
        ptr = strchr (buffy, '\n');
        *ptr = 0x0;
        ptr = strstr (buffy, "200");
        if (!ptr) {
	  failf (data,
		 "Error: Proxy did not return HTTP/1.0  200 OK header\n"
		 "Received: %s", buffy);
	  return URG_READ_ERROR;
        }
        infof (data, "Proxy replied to CONNECT request: %s\n",  buffy);

      }

      /* now, perform the SSL initialization for this socket */
      UrgSSLConnect (data);
  }

  now = time(NULL); /* time this *after* the connect is done */
  bytecount = 0;
  
  /* Figure out the ip-number and the first host name it shows: */
  {
    struct in_addr in;
    (void) memcpy(&in.s_addr, *hp->h_addr_list, sizeof (in.s_addr));
    infof(data, "Connected to %s (%s)\n", hp->h_name, inet_ntoa(in));
  }

  if((conf&(CONF_FTP|CONF_PROXY)) == CONF_FTP) {
    /* this is FTP and no proxy, we don't do the usual crap then */

    /* The first thing we do is wait for the "220*" line: */
    nread = GetLastResponse(data->firstsocket, buf, data);
    if(strncmp(buf, "220", 3)) {
      failf(data, "This doesn't seem like a nice ftp-server response");
      return URG_FTP_WEIRD_SERVER_REPLY;
    }

    /* send USER */
    sendf(data->firstsocket, data, "USER %s\n", ftpuser);

    /* wait for feedback */
    nread = GetLastResponse(data->firstsocket, buf, data);

    if(!strncmp(buf, "530", 3)) {
      /* 530 User ... access denied
         (the server denies to log the specified user) */
      failf(data, "Access denied: %s", &buf[4]);
      return URG_FTP_ACCESS_DENIED;
    }
    else if(!strncmp(buf, "331", 3)) {
      /* 331 Password required for ...
         (the server requires to send the user's password too) */
      sendf(data->firstsocket, data, "PASS %s\n", ftppasswd);
      nread = GetLastResponse(data->firstsocket, buf, data);

      if(!strncmp(buf, "530", 3)) {
        /* 530 Login incorrect.
           (the username and/or the password are incorrect) */
        failf(data, "the username and/or the password are incorrect");
        return URG_FTP_USER_PASSWORD_INCORRECT;
      }
      else if(!strncmp(buf, "230", 3)) {
        /* 230 User ... logged in.
           (user successfully logged in) */
        
        infof(data, "We have successfully logged in\n");
      }
      else {
        failf(data, "Odd return code after PASS");
        return URG_FTP_WEIRD_PASS_REPLY;
      }
    }
    else if(! strncmp(buf, "230", 3)) {
      /* 230 User ... logged in.
	(the user logged in without password) */
      infof(data, "We have successfully logged in\n");
    }
    else {
      failf(data, "Odd return code after USER");
      return URG_FTP_WEIRD_USER_REPLY;
    }

    if(data->conf & CONF_FTPPORT) {
#define FTPPORT 4012 /* start trying this port number */

      struct sockaddr_in sa;
      struct hostent *h=NULL;
      size_t size;
      unsigned short porttouse = FTPPORT+time(NULL)%5000;
      int i;
      char *myhost;
      
      if(data->ftpport) {
        myhost = if2ip(data->ftpport);
        if(myhost) {
          h = GetHost(data, myhost);
        }
        else {
          if(strlen(data->ftpport)>1)
            h = GetHost(data, data->ftpport);
          if(h)
            myhost=data->ftpport;
        }
      }
      if(!myhost) {
        myhost = getmyhost();
        h=GetHost(data, myhost);
      }
      infof(data, "We connect from %s\n", myhost);

      if ( h ) {
	if( (portsock = socket(AF_INET, SOCK_STREAM, 0)) >= 0 ) {
	  memset((char *)&sa, 0, sizeof(sa));
	  memcpy((char *)&sa.sin_addr, h->h_addr, h->h_length);
	  sa.sin_family = AF_INET;
	  sa.sin_addr.s_addr = INADDR_ANY;
	  sa.sin_port = 0;
	  size = sizeof(sa);

          if(bind(portsock, (struct sockaddr *)&sa, size) >= 0) {
	    /* we succeeded to bind */
            struct sockaddr_in add;
            size = sizeof(add);

            if(getsockname(portsock, (struct sockaddr *) &add, &size)<0) {
              failf(data, "getsockname() failed");
              return URG_FTP_PORT_FAILED;
            }
            porttouse = ntohs(add.sin_port);

	    if ( listen(portsock, 1) < 0 ) {
	      failf(data, "listen(2) failed on socket");
	      return URG_FTP_PORT_FAILED;
	    }
	  }
	  else {
	    failf(data, "bind(2) failed on socket");
	    return URG_FTP_PORT_FAILED;
	  }
	}
	else {
	  failf(data, "socket(2) failed (%s)");
	  return URG_FTP_PORT_FAILED;
	}
      }
      else {
	failf(data, "could't find my own IP address (%s)", myhost);
	return URG_FTP_PORT_FAILED;
      }
      {
	struct in_addr in;
	unsigned short ip[5];
	(void) memcpy(&in.s_addr, *h->h_addr_list, sizeof (in.s_addr));
	sscanf( inet_ntoa(in), "%hu.%hu.%hu.%hu",
		&ip[0], &ip[1], &ip[2], &ip[3]);
	sendf(data->firstsocket, data, "PORT %d,%d,%d,%d,%d,%d\n",
	      ip[0], ip[1], ip[2], ip[3],
	      porttouse >> 8,
	      porttouse & 255);
      }

      nread = GetLastResponse(data->firstsocket, buf, data);

      if(strncmp(buf, "200", 3)) {
	failf(data, "Server does not grok PORT, try without it!");
	return URG_FTP_PORT_FAILED;
      }     
    }
    else {

      sendf(data->firstsocket, data, "PASV\n");

      nread = GetLastResponse(data->firstsocket, buf, data);

      if(strncmp(buf, "227", 3)) {
	failf(data, "Odd return code after PASV");
	return URG_FTP_WEIRD_PASV_REPLY;
      }
      else {
	int ip[4];
	int port[2];
	unsigned short newport;
	char newhost[32];
	struct hostent *he;
	/* 227 Entering Passive Mode (127,0,0,1,4,51) */
	if(6 != sscanf(buf, "%*[^(](%d,%d,%d,%d,%d,%d)",
		       &ip[0], &ip[1], &ip[2], &ip[3],
		       &port[0], &port[1])) {
	  failf(data, "Oddly formatted 227-reply: %s", buf);
	  return URG_FTP_WEIRD_227_FORMAT;
	}
	sprintf(newhost, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
	he = GetHost(data, newhost);
	if(!he) {
	  failf(data, "Can't resolve new host %s", newhost);
	  return URG_FTP_CANT_GET_HOST;
	}

	newport = (port[0]<<8) + port[1];
	data->secondarysocket = socket(AF_INET, SOCK_STREAM, 0);

	memset((char *) &serv_addr, '\0', sizeof(serv_addr));
	memcpy((char *)&(serv_addr.sin_addr), he->h_addr, he->h_length);
	serv_addr.sin_family = he->h_addrtype;
	serv_addr.sin_port = htons(newport);
	
	if (connect(data->secondarysocket, (struct sockaddr *) &serv_addr,
		    sizeof(serv_addr)) < 0) {
	  switch(errno) {
#ifndef WIN32
	    /* this should be made nicer */
	  case ECONNREFUSED:
	    failf(data, "Connection refused");
	    break;
	  case EINTR:
	    failf(data, "Connection timeouted");
	    break;
#endif
	  default:
	    failf(data, "Can't connect to server");
	    break;
	  }
	  return URG_FTP_CANT_RECONNECT;
	}
      }

    }
    /* we have the (new) data connection ready */

    if(conf & CONF_UPLOAD) {
      /* Set type to binary */
      sendf(data->firstsocket, data, "TYPE I\n");

      nread = GetLastResponse(data->firstsocket, buf, data);

      if(strncmp(buf, "200", 3)) {
	failf(data, "Couldn't set binary mode");
	return URG_FTP_COULDNT_SET_BINARY;
      }

      if(data->resume_from) {
	/* we're about to continue the uploading of a file */
	/* 1. get already existing file's size. We use the SIZE
	      command for this which may not exist in the server!
	      The SIZE command is not in RFC959. */

	/* 2. This used to set REST. But since we can do append, we
    	      don't another ftp command. We just skip the source file
	      offset and then we APPEND the rest on the file instead */

	/* 3. pass file-size number of bytes in the source file */
	/* 4. lower the infilesize counter */
	/* => transfer as usual */

	if(data->resume_from < 0 ) {
	  /* we could've got a specified offset from the command line,
	     but now we know we didn't */

	  sendf(data->firstsocket, data, "SIZE %s\n", ppath);

	  nread = GetLastResponse(data->firstsocket, buf, data);

	  if(strncmp(buf, "213", 3)) {
	    failf(data, "Couldn't get file size: %s", buf+4);
	    return URG_FTP_COULDNT_GET_SIZE;
	  }

	  /* get the size from the ascii string: */
	  data->resume_from = atoi(buf+4);
	}

	if(data->resume_from) {
	  /* do we still game? */
	  int passed=0;
#if 0
	  /* Set resume file transfer offset */
	  infof(data, "Instructs server to resume from offset %d\n",
		data->resume_from);

	  sendf(data->firstsocket, data, "REST %d\n", data->resume_from);

	  nread = GetLastResponse(data->firstsocket, buf, data);

	  if(strncmp(buf, "350", 3)) {
	    failf(data, "Couldn't use REST: %s", buf+4);
	    return URG_FTP_COULDNT_USE_REST;
	  }
#else
	  /* enable append instead */
	  data->conf |= CONF_FTPAPPEND;
#endif
	  /* Now, let's read off the proper amount of bytes from the
	     input. If we knew it was a proper file we could've just
	     fseek()ed but we only have a stream here */
	  do {
	    int readthisamountnow = (data->resume_from - passed);
	    int actuallyread;

	    if(readthisamountnow > BUFSIZE)
	      readthisamountnow = BUFSIZE;

	    actuallyread =
	      data->fread(data->buffer, 1, readthisamountnow, data->in);

	    passed += actuallyread;
	    if(actuallyread != readthisamountnow) {
	      failf(data, "Could only read %d bytes from the input\n",
		    passed);
	      return URG_FTP_COULDNT_USE_REST;
	    }
	  }
	  while(passed != data->resume_from);

	  /* now, decrease the size of the read */
	  if(data->infilesize>0) {
	    data->infilesize -= data->resume_from;

            if(data->infilesize <= 0) {
              infof(data, "File already completely uploaded\n");
              return URG_OK;
            }
          }
	  /* we've passed, proceed as normal */
	}
      }

      /* Send everything on data->in to the socket */
      if(data->conf & CONF_FTPAPPEND)
        /* we append onto the file instead of rewriting it */
        sendf(data->firstsocket, data, "APPE %s\n", ppath);
      else
        sendf(data->firstsocket, data, "STOR %s\n", ppath);

      nread = GetLastResponse(data->firstsocket, buf, data);

      if(atoi(buf)>=400) {
	failf(data, "Failed FTP upload:%s", buf+3);
	/* oops, we never close the sockets! */
	return URG_FTP_COULDNT_STOR_FILE;
      }

      if(data->conf & CONF_FTPPORT) {
	result = AllowServerConnect(data, portsock);
	if( result )
	  return result;
      }

      bytecount=0;

      /* When we know we're uploading a specified file, we can get the file
	 size prior to the actual upload. */

      ProgressInit(data, data->infilesize);
      result = Upload(data, data->secondarysocket, data->infilesize,
		      &bytecount);
      if(result)
	return result;
      
      if((-1 != data->infilesize) && (data->infilesize != bytecount)) {
	failf(data, "Wrote only partial file (%d out of %d bytes)",
	      bytecount, data->infilesize);
	return URG_FTP_PARTIAL_FILE;
      }
    }
    else {
      /* Retrieve file or directory */
      bool dirlist=FALSE;

      if(!ppath[0])
          /* make sure this becomes a valid name */
          ppath="./";

      if('/' == ppath[strlen(ppath)-1]) {
	/* The specified path ends with a slash, and therefore we think this
	   is a directory that is requested, use LIST. But before that we
	   need to set ASCII transfer mode. */
	dirlist = TRUE;

	/* Set type to ASCII */
	sendf(data->firstsocket, data, "TYPE A\n");
	
	nread = GetLastResponse(data->firstsocket, buf, data);
	
	if(strncmp(buf, "200", 3)) {
	  failf(data, "Couldn't set ascii mode");
	  return URG_FTP_COULDNT_SET_ASCII;
	}

	/* if this output is to be machine-parsed, the NLST command will be
	   better used since the LIST command output is not specified or
	   standard in any way */

	sendf(data->firstsocket, data, "%s %s\n",
	      data->conf&CONF_FTPLISTONLY?"NLST":"LIST",
	      ppath);
      }
      else {
	/* Set type to binary */
	sendf(data->firstsocket, data, "TYPE I\n");

	nread = GetLastResponse(data->firstsocket, buf, data);

	if(strncmp(buf, "200", 3)) {
	  failf(data, "Couldn't set binary mode");
	  return URG_FTP_COULDNT_SET_BINARY;
	}

	if(data->resume_from) {
	  /* Set resume file transfer offset */
	  infof(data, "Instructs server to resume from offset %d\n",
		data->resume_from);

	  sendf(data->firstsocket, data, "REST %d\n", data->resume_from);

	  nread = GetLastResponse(data->firstsocket, buf, data);

	  if(strncmp(buf, "350", 3)) {
	    failf(data, "Couldn't use REST: %s", buf+4);
	    return URG_FTP_COULDNT_USE_REST;
	  }
	}

	sendf(data->firstsocket, data, "RETR %s\n", ppath);
      }
      nread = GetLastResponse(data->firstsocket, buf, data);

      if(!strncmp(buf, "150", 3)) {
	/* 150 Opening BINARY mode data connection for /etc/passwd (2241
           bytes).  (ok, the file is being transfered) */
	
	/* 150 Opening ASCII mode data connection for /bin/ls */

	/* 150 ASCII data connection for /bin/ls (137.167.104.91,37445) (0 bytes). */

	int size=-1; /* default unknown size */

	if(!dirlist)
	  /* it seems directory listings either don't show the size or very
	     often uses size 0 anyway */
	  sscanf(buf, "%*[^(](%d", &size);

	if((size > -1) && (data->resume_from>=0)) {
	  size -= data->resume_from;
	  if(size <= 0) {
	    failf(data, "Offset (%d) was beyond file size (%d)",
		  data->resume_from, data->resume_from+size);
	    return URG_FTP_PARTIAL_FILE;
	  }
	}

	if(data->conf & CONF_FTPPORT) {
	  result = AllowServerConnect(data, portsock);
	  if( result )
	    return result;
	}

	ProgressInit(data, size);

	infof(data, "Getting file with size: %d\n", size);

	result=Download(data, data->secondarysocket, size, FALSE,
			&bytecount);
	if(result)
	  return result;

	if((-1 != size) && (size != bytecount)) {
	  failf(data, "Received only partial file");
	  return URG_FTP_PARTIAL_FILE;
	}
	else if(0 == bytecount) {
	  failf(data, "No data was received!");
	  return URG_FTP_COULDNT_RETR_FILE;
	}
      }
      else {
	failf(data, "%s", buf+4);
	return URG_FTP_COULDNT_RETR_FILE;
      }
	
    }
    /* end of transfer */
    ProgressEnd(data);

    /* shut down the socket to inform the server we're done */
    sclose(data->secondarysocket);
    data->secondarysocket = -1;
    
    /* now let's see what the server says about the transfer we
       just performed: */
    nread = GetLastResponse(data->firstsocket, buf, data);

    /* 226 Transfer complete */
    if(strncmp(buf, "226", 3)) {
      failf(data, "%s", buf+4);
      return URG_FTP_WRITE_ERROR;
    }
  } else

  /* upload raw http data, headers and all */
  if((conf&(CONF_HTTP|CONF_UPLOAD)) == (CONF_HTTP|CONF_UPLOAD)) {
    ProgressInit(data, data->infilesize);
    result = Upload(data, data->firstsocket, data->infilesize,
                    &bytecount);
    if(result)
      return result;

    if((-1 != data->infilesize) && (data->infilesize != bytecount)) {
      failf(data, "Wrote only partial POST (%d out of %d bytes)",
            bytecount, data->infilesize);
/*
      return URG_FTP_PARTIAL_FILE;
*/
    }

    result = Download(data, data->firstsocket, -1, TRUE, &bytecount);
    if(result)
      return result;

    ProgressEnd(data);

  } else 


  {
    /* Send the GET line to the HTTP server */

    char userpwd[512];
    char rangeline[512];
    char content[80];
    char ref[512];

    if(conf & CONF_USERPWD) {
      char authorization[512];
      sprintf(userpwd, "%s:%s", ftpuser, ftppasswd);
      base64Encode(userpwd, authorization);
      sprintf(userpwd,
              "Authorization: Basic %s\015\012",
              authorization);
    }
    if(conf & CONF_RANGE) {
      sprintf(rangeline, "Range: bytes=%s\015\012", data->range);
    }
    if(conf & CONF_POST) {
      sprintf(content, "Content-length: %d\015\012",
              strlen(data->postfields)) ;
    }
    if(conf & CONF_REFERER) {
      sprintf(ref, "Referer: %s\015\012", data->referer);
    }

    if ((conf & CONF_PROXY && (!(conf & CONF_HTTPS))))  {
      /* The path the is sent to the proxy is in fact the entire URL */
      strncpy(ppath, data->url, 511);
    }

    sendf(data->firstsocket, data,
          "%s %s HTTP/1.0\015\012"
          "%s"
          "%s"
          "%s"
          "%s"
          "%s"
          "Host: %s\015\012"
          "Pragma: no-cache\015\012"
          "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, */*\015\012"
          "%s"
          "%s%s"
          "%s\015\012",

          conf&CONF_NOBODY?"HEAD":(conf&CONF_POST?"POST":"GET"),
          ppath,
          (conf&CONF_PROXYUSERPWD)?proxyuserpwd:"",
          (conf&CONF_USERPWD)?userpwd:"",
          (conf&CONF_KEEPALIVE)?"Connection: Keep-Alive\015\012":"",
          (conf&CONF_RANGE)?rangeline:"",
          (data->useragent)?uagent:"",
          name, /* host */
	  (conf&CONF_REFERER)?ref:"",
          (conf&CONF_POST)?content:"",
          (conf&CONF_POST)?"Content-type: application/x-www-form-urlencoded\015\012\015\012":"",
          (conf&CONF_POST)?data->postfields:""
          );

    result = Download(data, data->firstsocket, -1, TRUE, &bytecount);
    if(result)
      return result;

    ProgressEnd(data);

  }
  if(bytecount) {
    time_t end=time(NULL);
    infof(data, "%i bytes transfered in %d seconds (%d bytes/sec).\n",
          bytecount, end-now, bytecount/(end-now?end-now:1));
  }
  return URG_OK;
}

/* infof() is for info message along the way */

static void infof(struct UrlData *data, char *fmt, ...)
{
  va_list ap;
  if(data->conf & CONF_VERBOSE) {
    va_start(ap, fmt);
    fputs("* ", stderr);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
  }
}

/* failf() is for messages stating why we failed, the LAST one will be
   returned for the user (if requested) */

static void failf(struct UrlData *data, char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  if(data->errorbuffer)
    vsprintf(data->errorbuffer, fmt, ap);
  else /* no errorbuffer receives this, write to stderr instead */
    vfprintf(stderr, fmt, ap);
  va_end(ap);
}

/* sendf() sends the formated data to the server */

static int sendf(int fd, struct UrlData *data, char *fmt, ...)
{
  size_t bytes_written;
  char s[1024];
  va_list ap;
  va_start(ap, fmt);
  vsprintf(s, fmt, ap);
  va_end(ap);
  if(data->conf & CONF_VERBOSE)
    fprintf(stderr, "> %s", s);
#ifndef USE_SSLEAY
   bytes_written = swrite(fd, s, strlen(s));
#else
  if (data->use_ssl) {
    bytes_written = SSL_write(data->ssl, s, strlen(s));
  } else {
    bytes_written = swrite(fd, s, strlen(s));
  }
#endif /* USE_SSLEAY */
  return(bytes_written);
}
